# $Id: control.rb,v 1.32 2004/10/03 11:53:18 toki Exp $

require 'forwardable'
require 'rucy/params'

module Rucy
  class Control
    extend Forwardable

    def initialize
      @loader = nil
      @pw_enc = PasswordEncryptor.new
      @doc_errors = Array.new
      @filter_errors = Array.new
      @logfiles_output = Array.new
      @logging_errors = Array.new
      @alias_errors = Array.new
      @restart_signal = nil
      @properties = nil
      @page_path = nil
      @server = nil
    end

    attr_writer :loader
    attr_writer :pw_enc
    attr_writer :properties
    attr_writer :page_path
    attr_writer :restart_signal

    def_delegator :@loader, :load_errors
    attr_reader :doc_errors
    attr_reader :filter_errors
    attr_reader :logging_errors
    attr_reader :alias_errors

    def add_logfile(output)
      @logfiles_output.push(output)
      nil
    end

    def server_setup(server)
      logging_params = self.logging_params
      logging_params.server_setup(self, server)

      access_log_params = self.access_log_params
      access_log_params.server_setup(self, server)

      @loader.load
      for mod_path, error in @loader.load_errors
	server.err("error: not loaded #{mod_path.inspect}: #{error}")
      end
      doc_factory = @loader.factory
      doc_factory.add_option(:pw_enc, @pw_enc)
      doc_factory.setup

      server_params = self.server_params
      server_params.server_setup(self, server)

      alias_list = self.aliases
      alias_list.server_setup(self, server)

      document_list = self.documents
      for mount_info in document_list
	name = mount_info['document']
	args = mount_info['arguments']
	path = mount_info['mount_path']
	mask = mount_info['mount_mask']
	host = mount_info['virtual_host']
	begin
	  document = doc_factory.doc_build(name, args)
	  if (host) then
	    host += ':' + server_params.port.to_s
	    server.virtual_mount(host, document, path, mask)
	  else
	    server.mount(document, path, mask)
	  end
	rescue StandardError, ScriptError
	  server.err("error: not mounted #{name.inspect} document: (#{$!.class}) #{$!}")
	  @doc_errors.push({ :document => name,
			     :arguments => args,
			     :mount_path => path,
			     :mount_mask => mask,
			     :virtual_host => host,
			     :exception => $!
			   })
	end
      end

      filter_list = self.filters
      for filter_info in filter_list
	name = filter_info['filter']
	args = filter_info['arguments']
	path = filter_info['attach_path']
	mask = filter_info['attach_mask']
	host = filter_info['virtual_host']
	begin
	  filter = doc_factory.filter_build(name, args)
	  if (host) then
	    host += ':' + server_params.port.to_s
	    server.virtual_attach(host, filter, path, mask)
	  else
	    server.attach(filter, path, mask)
	  end
	rescue StandardError, ScriptError
	  server.err("error: not attached #{name.inspect} filter: (#{$!.class}) #{$!}")
	  @filter_errors.push({ :filter => name,
				:arguments => args,
				:attach_path => path,
				:attach_mask => mask,
				:virtual_host => host,
				:exception => $!
			      })
	end
      end

      admin_params = self.admin_params
      admin_params.server_setup(self, server, @loader, doc_factory, @pw_enc, @page_path)

      @server = server
      server.close_hook{|s|
	logfiles_close
      }

      nil
    end

    def server_restart
      @restart_signal.notify_restart
      nil
    end

    def server_close
      @restart_signal.notify_close
      nil
    end

    def logfiles_close
      for output in @logfiles_output
	output.close
	@server.debug("close #{output.path.inspect}.")
      end
      nil
    end

    def admin_params
      AdminParams.load(@properties)
    end

    def set_admin_params(params)
      case (params)
      when AdminParams
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def server_params
      ServerParams.load(@properties)
    end

    def set_server_params(params)
      case (params)
      when ServerParams
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def logging_params
      LoggingParams.load(@properties)
    end

    def set_logging_params(params)
      case (params)
      when LoggingParams
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def access_log_params
      AccessLogParams.load(@properties)
    end

    def set_access_log_params(params)
      case (params)
      when AccessLogParams
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def aliases
      AliasList.load(@properties)
    end

    def set_aliases(alias_list)
      case (alias_list)
      when AliasList
	alias_list.save(@properties)
      else
	raise "unknown parameter set: #{alias_list.inspect}"
      end
      nil
    end

    def documents
      @properties.list('documents')
    end

    def set_documents(document_list)
      @properties.set_list('documents', document_list)
      nil
    end

    def filters
      @properties.list('filters')
    end

    def set_filters(filter_list)
      @properties.set_list('filters', filter_list)
      nil
    end
  end
end
